import json
from typing import List
from unittest.mock import Mock, mock_open, patch

import click
import pytest

from anyscale.controllers.cluster_controller import ClusterController
from anyscale.sdk.anyscale_client import StartClusterOptions


@pytest.mark.parametrize(
    "cluster",
    [
        Mock(state="Running", cluster_environment_build_id="build1"),
        Mock(state="Terminated", cluster_environment_build_id="build1"),
    ],
)
@pytest.mark.parametrize(
    "build", [Mock(id="build1", revision=1), Mock(id="build2", revision=2)]
)
@pytest.mark.parametrize("cluster_compute", [Mock(id="compute1"), Mock(id="compute2")])
@pytest.mark.parametrize("passed_cluster_env", [True, False])
@pytest.mark.parametrize("passed_cluster_compute", [True, False])
def test_check_needs_start(
    cluster: Mock,
    build: Mock,
    cluster_compute: Mock,
    passed_cluster_env: bool,
    passed_cluster_compute: bool,
    mock_auth_api_client,
) -> None:
    cluster_controller = ClusterController()

    needs_start = cluster_controller._check_needs_start(
        cluster, build, cluster_compute, passed_cluster_env, passed_cluster_compute
    )
    if cluster.state != "Running":
        assert needs_start
    elif not passed_cluster_env and not passed_cluster_compute:
        assert not needs_start
    elif passed_cluster_env and cluster.cluster_environment_build_id != build.id:
        assert needs_start
    elif passed_cluster_compute and cluster.cluster_compute_id != cluster_compute.id:
        assert needs_start


def test_get_project_id_and_cluster_name_given_cluster_id(mock_auth_api_client):
    cluster_controller = ClusterController()

    # Test passing in cluster id only
    mock_cluster = Mock(id="mock_cluster_id", project_id="mock_project_id")
    mock_cluster.name = "mock_cluster_name"
    cluster_controller.anyscale_api_client.get_cluster = Mock(
        return_value=Mock(result=mock_cluster)
    )
    assert cluster_controller._get_project_id_and_cluster_name(
        cluster_id="mock_cluster_id",
        project_id=None,
        cluster_name=None,
        project_name=None,
    ) == ("mock_project_id", "mock_cluster_name")

    # Test passing in project id
    cluster_controller._get_or_generate_cluster_name = Mock(  # type: ignore
        return_value="mock_cluster_name1"
    )
    mock_get_and_validate_project_id = Mock(return_value="mock_project_id1")
    with patch.multiple(
        "anyscale.controllers.cluster_controller",
        get_and_validate_project_id=mock_get_and_validate_project_id,
    ):
        assert cluster_controller._get_project_id_and_cluster_name(
            cluster_id=None,
            project_id="mock_project_id1",
            cluster_name=None,
            project_name=None,
        ) == ("mock_project_id1", "mock_cluster_name1")
        mock_get_and_validate_project_id.assert_called_once_with(
            project_id="mock_project_id1",
            project_name=None,
            api_client=cluster_controller.api_client,
            anyscale_api_client=cluster_controller.anyscale_api_client,
        )

    # Test passing in project name
    cluster_controller._get_or_generate_cluster_name = Mock(  # type: ignore
        return_value="mock_cluster_name1"
    )
    mock_get_and_validate_project_id.reset_mock()
    with patch.multiple(
        "anyscale.controllers.cluster_controller",
        get_and_validate_project_id=mock_get_and_validate_project_id,
    ):
        assert cluster_controller._get_project_id_and_cluster_name(
            cluster_id=None,
            project_id=None,
            cluster_name=None,
            project_name="mock_project_name1",
        ) == ("mock_project_id1", "mock_cluster_name1")
        mock_get_and_validate_project_id.assert_called_once_with(
            project_id=None,
            project_name="mock_project_name1",
            api_client=cluster_controller.api_client,
            anyscale_api_client=cluster_controller.anyscale_api_client,
        )


def test_get_cluster_env_and_build(mock_auth_api_client):
    cluster_controller = ClusterController()
    mock_get_build_from_cluster_env_identifier = Mock(return_value=Mock(id="build_id1"))
    mock_get_default_cluster_env_build = Mock(return_value=Mock(id="default_build_id"))

    with patch.multiple(
        "anyscale.controllers.cluster_controller",
        get_build_from_cluster_env_identifier=mock_get_build_from_cluster_env_identifier,
        get_default_cluster_env_build=mock_get_default_cluster_env_build,
    ):
        assert (
            cluster_controller._get_cluster_env_and_build(None)[1].id
            == "default_build_id"
        )
        assert (
            cluster_controller._get_cluster_env_and_build("mock_cluster_env")[1].id
            == "build_id1"
        )


def test_get_cluster_compute(mock_auth_api_client):
    cluster_controller = ClusterController()
    mock_get_cluster_compute_from_name = Mock(return_value=Mock(id="cluster_compute1"))
    cluster_controller.anyscale_api_client.create_cluster_compute = Mock(
        return_value=Mock(result=Mock(id="cluster_compute2"))
    )
    mock_get_default_cluster_compute = Mock(
        return_value=Mock(id="default_cluster_compute")
    )

    with patch(
        "builtins.open",
        mock_open(
            read_data=json.dumps(
                {
                    "cloud_id": "mock_cloud_id",
                    "region": "mock_region",
                    "head_node_type": "mock_head_node_type",
                    "worker_node_types": "mock_worker_node_types",
                }
            )
        ),
    ), patch.multiple(
        "anyscale.controllers.cluster_controller",
        get_cluster_compute_from_name=mock_get_cluster_compute_from_name,
        get_default_cluster_compute=mock_get_default_cluster_compute,
    ):
        assert (
            cluster_controller._get_cluster_compute(
                "cluster_compute_name1", None, None, "mock_project_id"
            ).id
            == "cluster_compute1"
        )
        assert (
            cluster_controller._get_cluster_compute(
                None, "cluster_compute_file2", None, "mock_project_id"
            ).id
            == "cluster_compute2"
        )
        assert (
            cluster_controller._get_cluster_compute(
                None, None, None, "mock_project_id"
            ).id
            == "default_cluster_compute"
        )


@pytest.mark.parametrize("needs_start", [True, False])
def test_start(mock_auth_api_client, needs_start: bool) -> None:
    cluster_controller = ClusterController()
    mock_build = Mock(id="build_id")
    mock_cluster_compute = Mock(id="cluster_compute_id")
    cluster_controller._get_project_id_and_cluster_name = Mock(  # type: ignore
        return_value=("mock_project_id", "mock_cluster_name")
    )
    cluster_controller._get_cluster_env_and_build = Mock(  # type: ignore
        return_value=(Mock(id="cluster_env_id"), mock_build)
    )
    cluster_controller._get_cluster_compute = Mock(return_value=mock_cluster_compute)  # type: ignore
    cluster_controller._create_or_update_cluster_data = Mock(  # type: ignore
        return_value=(Mock(id="cluster_id"), needs_start)
    )
    mock_user_service_access = "private"

    with patch.multiple(
        "anyscale.controllers.cluster_controller", wait_for_session_start=Mock(),
    ):
        cluster_controller.start(
            cluster_name=None,
            cluster_id="cluster_id",
            cluster_env_name="cluster_env_name",
            docker=None,
            python_version=None,
            ray_version=None,
            cluster_compute_name="cluster_compute_name",
            cluster_compute_file=None,
            cloud_name=None,
            idle_timeout=None,
            project_id=None,
            project_name=None,
            user_service_access=mock_user_service_access,
        )

    cluster_controller._get_project_id_and_cluster_name.assert_called_once_with(
        "cluster_id", None, None, None
    )
    cluster_controller._get_cluster_env_and_build.assert_called_once_with(
        "cluster_env_name"
    )
    cluster_controller._get_cluster_compute.assert_called_once_with(
        "cluster_compute_name", None, None, "mock_project_id"
    )
    cluster_controller._create_or_update_cluster_data.assert_called_once_with(
        "mock_cluster_name",
        "mock_project_id",
        mock_build,
        mock_cluster_compute,
        True,
        True,
        None,
        mock_user_service_access,
    )

    if needs_start:
        cluster_controller.anyscale_api_client.start_cluster.assert_called_once_with(
            "cluster_id",
            StartClusterOptions(
                cluster_environment_build_id=mock_build.id,
                cluster_compute_id=mock_cluster_compute.id,
            ),
        )


@pytest.mark.parametrize("cluster_list", [[], [Mock(id="mock_cluster_id")]])
def test_terminate(mock_auth_api_client, cluster_list: List[Mock],) -> None:
    cluster_controller = ClusterController()
    cluster_controller._get_project_id_and_cluster_name = Mock(  # type: ignore
        return_value=("mock_project_id", "mock_cluster_name")
    )
    cluster_controller.anyscale_api_client.search_clusters = Mock(  # type: ignore
        return_value=Mock(results=cluster_list)
    )

    with pytest.raises(click.ClickException):
        cluster_controller.terminate(None, None, None, None)
    with pytest.raises(click.ClickException):
        cluster_controller.terminate("cluster_name", "cluster_id", None, None)
    if len(cluster_list) == 0:
        with pytest.raises(click.ClickException):
            cluster_controller.terminate("cluster_name", None, "project_id", None)
    else:
        cluster_controller.terminate("cluster_name", None, "project_id", None)
        cluster_controller._get_project_id_and_cluster_name.assert_called_once_with(
            None, "project_id", "cluster_name", None
        )
        cluster_controller.anyscale_api_client.search_clusters.assert_called_once_with(
            {"project_id": "mock_project_id", "name": {"equals": "mock_cluster_name"}}
        )
        cluster_controller.anyscale_api_client.terminate_cluster.assert_called_once_with(
            "mock_cluster_id", {}
        )


@pytest.mark.parametrize("cluster_list", [[], [Mock(id="mock_cluster_id")]])
def test_archive(mock_auth_api_client, cluster_list: List[Mock],) -> None:
    cluster_controller = ClusterController()
    cluster_controller._get_project_id_and_cluster_name = Mock(  # type: ignore
        return_value=("mock_project_id", "mock_cluster_name")
    )
    cluster_controller.anyscale_api_client.search_clusters = Mock(  # type: ignore
        return_value=Mock(results=cluster_list)
    )

    with pytest.raises(click.ClickException):
        cluster_controller.archive(None, None, None, None)
    with pytest.raises(click.ClickException):
        cluster_controller.archive("cluster_name", "cluster_id", None, None)
    if len(cluster_list) == 0:
        with pytest.raises(click.ClickException):
            cluster_controller.archive("cluster_name", None, "project_id", None)
    else:
        cluster_controller.archive("cluster_name", None, "project_id", None)
        cluster_controller._get_project_id_and_cluster_name.assert_called_once_with(
            None, "project_id", "cluster_name", None
        )
        cluster_controller.anyscale_api_client.search_clusters.assert_called_once_with(
            {"project_id": "mock_project_id", "name": {"equals": "mock_cluster_name"}}
        )
        cluster_controller.anyscale_api_client.archive_cluster.assert_called_once_with(
            "mock_cluster_id"
        )



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class LogFilter(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'glob': 'str',
        'node_id': 'str',
        'job_id': 'str',
        'job_run': 'str',
        'node_type': 'NodeType'
    }

    attribute_map = {
        'glob': 'glob',
        'node_id': 'node_id',
        'job_id': 'job_id',
        'job_run': 'job_run',
        'node_type': 'node_type'
    }

    def __init__(self, glob='**', node_id=None, job_id=None, job_run=None, node_type=None, local_vars_configuration=None):  # noqa: E501
        """LogFilter - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._glob = None
        self._node_id = None
        self._job_id = None
        self._job_run = None
        self._node_type = None
        self.discriminator = None

        if glob is not None:
            self.glob = glob
        if node_id is not None:
            self.node_id = node_id
        if job_id is not None:
            self.job_id = job_id
        if job_run is not None:
            self.job_run = job_run
        if node_type is not None:
            self.node_type = node_type

    @property
    def glob(self):
        """Gets the glob of this LogFilter.  # noqa: E501

        glob filter for selecting logs, default **.  # noqa: E501

        :return: The glob of this LogFilter.  # noqa: E501
        :rtype: str
        """
        return self._glob

    @glob.setter
    def glob(self, glob):
        """Sets the glob of this LogFilter.

        glob filter for selecting logs, default **.  # noqa: E501

        :param glob: The glob of this LogFilter.  # noqa: E501
        :type: str
        """

        self._glob = glob

    @property
    def node_id(self):
        """Gets the node_id of this LogFilter.  # noqa: E501

        node id to download logs, default None.  # noqa: E501

        :return: The node_id of this LogFilter.  # noqa: E501
        :rtype: str
        """
        return self._node_id

    @node_id.setter
    def node_id(self, node_id):
        """Sets the node_id of this LogFilter.

        node id to download logs, default None.  # noqa: E501

        :param node_id: The node_id of this LogFilter.  # noqa: E501
        :type: str
        """

        self._node_id = node_id

    @property
    def job_id(self):
        """Gets the job_id of this LogFilter.  # noqa: E501

        job id to download logs, default None.  # noqa: E501

        :return: The job_id of this LogFilter.  # noqa: E501
        :rtype: str
        """
        return self._job_id

    @job_id.setter
    def job_id(self, job_id):
        """Sets the job_id of this LogFilter.

        job id to download logs, default None.  # noqa: E501

        :param job_id: The job_id of this LogFilter.  # noqa: E501
        :type: str
        """

        self._job_id = job_id

    @property
    def job_run(self):
        """Gets the job_run of this LogFilter.  # noqa: E501

        job run name to download logs, default None.  # noqa: E501

        :return: The job_run of this LogFilter.  # noqa: E501
        :rtype: str
        """
        return self._job_run

    @job_run.setter
    def job_run(self, job_run):
        """Sets the job_run of this LogFilter.

        job run name to download logs, default None.  # noqa: E501

        :param job_run: The job_run of this LogFilter.  # noqa: E501
        :type: str
        """

        self._job_run = job_run

    @property
    def node_type(self):
        """Gets the node_type of this LogFilter.  # noqa: E501

        node type to download logs, default None.  # noqa: E501

        :return: The node_type of this LogFilter.  # noqa: E501
        :rtype: NodeType
        """
        return self._node_type

    @node_type.setter
    def node_type(self, node_type):
        """Sets the node_type of this LogFilter.

        node type to download logs, default None.  # noqa: E501

        :param node_type: The node_type of this LogFilter.  # noqa: E501
        :type: NodeType
        """

        self._node_type = node_type

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, LogFilter):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, LogFilter):
            return True

        return self.to_dict() != other.to_dict()



from datetime import timedelta
from typing import Any, Dict, Optional

import click

from anyscale.cli_logger import BlockLogger
from anyscale.client.openapi_client.models.node_type import NodeType
from anyscale.controllers.session_controller import SessionController


log = BlockLogger()


@click.group(
    "logs",
    help="Download log and ray tune result files from Anyscale cluster or job.",
    hidden=True,
)
def log_cli() -> None:
    pass


@log_cli.command(
    name="file", help="Download log files of a cluster or job to BASE_FOLDER."
)
@click.argument("glob", type=str, required=False, default=None, envvar="GLOB")
@click.option("-c", "--cluster-id", type=str, default=None, help="ID of the cluster.")
@click.option("-j", "--job-id", type=str, default=None, help="ID of the job.")
@click.option("-r", "--job-run", type=str, default=None, help="Name of the job run.")
@click.option(
    "-n", "--node-id", type=str, default=None, help="Download logs of a specific node.",
)
@click.option(
    "--worker-only", is_flag=True, help="Download logs of only the worker nodes.",
)
@click.option(
    "--head-only", is_flag=True, help="Download logs of only the head node.",
)
@click.option(
    "--page-size",
    type=int,
    default=None,
    hidden=True,
    help="Results to get in a single API call.",
)
@click.option(
    "--ttl",
    type=int,
    default=None,
    hidden=True,
    help="TTL in seconds for the presigned URL.",
)
@click.option(
    "--parallelism",
    type=int,
    default=None,
    hidden=True,
    help="Number of files to download in parallel at a time.",
)
@click.option(
    "--read-timeout",
    type=int,
    default=None,
    hidden=True,
    help="Read timeout for the http call, while downloading files.",
)
@click.argument(
    "base_folder", type=str, required=True, default="", envvar="BASE_FOLDER"
)
def anyscale_download_log_files(
    glob: Optional[str],
    cluster_id: Optional[str],
    job_id: Optional[str],
    job_run: Optional[str],
    node_id: Optional[str],
    worker_only: bool,
    head_only: bool,
    page_size: Optional[int],
    ttl: Optional[int],
    parallelism: Optional[int],
    read_timeout: Optional[float],
    base_folder: str,
) -> None:
    """Download log files.
    """
    session_controller = SessionController()
    node_type: Optional[NodeType] = None

    if cluster_id is None and job_id is None and job_run is None:
        raise click.ClickException(
            "Atleast one of --cluster-id or --job-id or --job-run must be provided."
        )

    if worker_only and head_only:
        raise click.ClickException("Cannot specify both --worker-only and --head-only.")

    if worker_only:
        node_type = NodeType.WORKER_NODES
    elif head_only:
        node_type = NodeType.HEAD_NODE

    hidden_args: Dict[str, Any] = {"page_size": page_size, "ttl_seconds": ttl}
    if parallelism:
        hidden_args["parallelism"] = parallelism
    if read_timeout:
        hidden_args["read_timeout"] = timedelta(seconds=read_timeout)

    session_controller.download_log_files(
        base_folder=base_folder,
        glob=glob,
        session_id=cluster_id,
        node_id=node_id,
        job_id=job_id,
        job_run=job_run,
        node_type=node_type,
        **hidden_args
    )


@log_cli.command(
    name="results",
    help="Download ray tune result files of a cluster or job to BASE_FOLDER.",
)
@click.option("-c", "--cluster-id", type=str, default=None, help="ID of the cluster.")
@click.option("-j", "--job-id", type=str, default=None, help="ID of the job.")
@click.option("-r", "--job-run", type=str, default=None, help="Name of the job run.")
@click.option(
    "--parallelism",
    type=int,
    default=None,
    hidden=True,
    help="Number of files to download in parallel at a time.",
)
@click.option(
    "--read-timeout",
    type=int,
    default=None,
    hidden=True,
    help="Read timeout for the http call, while downloading files.",
)
@click.argument(
    "base_folder", type=str, required=True, default="", envvar="BASE_FOLDER"
)
def anyscale_download_ray_result_files(
    cluster_id: Optional[str],
    job_id: Optional[str],
    job_run: Optional[str],
    parallelism: Optional[int],
    read_timeout: Optional[float],
    base_folder: str,
) -> None:
    """Download ray tune result files.
    """
    session_controller = SessionController()

    if cluster_id is None and job_id is None and job_run is None:
        raise click.ClickException(
            "Atleast one of --cluster-id or --job-id or --job-run must be provided."
        )

    hidden_args: Dict[str, Any] = {}
    if parallelism:
        hidden_args["parallelism"] = parallelism
    if read_timeout:
        hidden_args["read_timeout"] = timedelta(seconds=read_timeout)

    session_controller.download_ray_result_files(
        base_folder=base_folder,
        session_id=cluster_id,
        job_id=job_id,
        job_run=job_run,
        **hidden_args
    )



from __future__ import absolute_import, division, print_function

import os
import re

from setuptools import Distribution, find_packages, setup


# mypy: ignore-errors


def find_version(path):
    with open(path) as f:
        match = re.search(
            r"^__version__ = ['\"]([^'\"]*)['\"]", f.read(), re.MULTILINE,
        )
        if match:
            return match.group(1)
        raise RuntimeError("Unable to find version string.")


def read_requirements(path):
    with open(path) as f:
        return [line.strip() for line in f.readlines()]


class BinaryDistribution(Distribution):
    def is_pure(self):
        return True

    def has_ext_modules(self):
        return True


# If adding new webterminal deps,
# Update backend/server/services/application_templates_service.py
# to prevent users from uninstalling them.
extras_require = {"backend": ["terminado==0.10.1", "tornado"]}


def package_files(directory):
    paths = []
    for (path, _, filenames) in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join("..", path, filename))
    return paths


VERSION_PATH = os.path.join(
    os.path.dirname(os.path.realpath(__file__)), "anyscale", "version.py"
)
REQUIREMENTS_PATH = os.path.join(
    os.path.dirname(os.path.realpath(__file__)), "requirements.in"
)

with open("README.md", "r") as fh:
    long_description = fh.read()


setup(
    name="anyscale",
    version=find_version(VERSION_PATH),
    author="Anyscale Inc.",
    description=("Command Line Interface for Anyscale"),
    long_description=long_description,
    packages=[*find_packages(exclude="tests")],
    distclass=BinaryDistribution,
    setup_requires=["setuptools_scm"],
    package_data={
        "": [
            "requirements.in",
            "*.yaml",
            "*.json",
            *package_files("anyscale/client"),
            *package_files("anyscale/sdk"),
        ],
    },
    install_requires=read_requirements(REQUIREMENTS_PATH),
    extras_require=extras_require,
    entry_points={"console_scripts": ["anyscale=anyscale.scripts:main"]},
    include_package_data=True,
    zip_safe=False,
)
